// Copyright (c) 2014, The Constrain project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library constraint;

import 'package:matcher/matcher.dart';
import 'package:constrain/metadata.dart';
import 'src/preconditions.dart';

class Constraint<T> {
  final ConstraintValidator f;
  final bool isOptional;
  const Constraint(this.f, {this.isOptional : true});

//  bool isValid(T value, ConstraintValidationContext context) {
//
//  }

//  bool isValid(T value, ConstraintValidationContext context) =>
//      validate(value, context).isEmpty;

  void validate(T value, ConstraintValidationContext context) {
    if (!isOptional && value == null) {
      context.addViolation(); // TODO => 'value must not be null'!!!
//      return new Set()..add(new ConstraintViolation('value must not be null'));
    }

    f(value, this, context);
  }
}

class Ensure extends Constraint<dynamic>{
  final dynamic matcher;
  const Ensure(this.matcher, {isOptional : true})
      : super(const _MatcherValidator(), isOptional: isOptional);

  void validate(dynamic value, ConstraintValidationContext context) {
    // TODO: ugly but not sure to get around making iNotNull !isOptional
    if (matcher == isNotNull) {
      f(value, this, context);
    }
    else {
      super.validate(value, context);
    }
  }

  String toString() => matcher.toString();
}

abstract class ConstraintValidationContext {
  void addViolation({String reason});
}

typedef void ConstraintValidator(fieldValue,
    Constraint constraint, ConstraintValidationContext context);



class ConstraintViolation<T> {
  final ConstraintDescriptor constraintDescriptor;
  final Object invalidValue;
  final Object leafObject;
  final Iterable<Symbol> propertyPath;
  final T rootObject;
  String _message;
  String reason;
  Type get rootType => rootObject == null ? null : rootObject.runtimeType;
  String get message {
    if (_message == null) {
      String objectDesc = rootType != null ?
          'Invalid $rootType ($rootObject)' : '';
      String path = propertyPath.join('.');
      String locationDesc = 'Constraint violated at path $path';
      // TODO: build up reason from non ensures
      // TODO: build proper message
      _message = '$objectDesc. $locationDesc\n$reason';
    }
    return _message;
  }

  ConstraintViolation(this.constraintDescriptor, this.invalidValue,
      this.leafObject, this.propertyPath, this.rootObject,
      {String message, this.reason})
      : this._message = message {

    ensure(constraintDescriptor, isNotNull);
    ensure(leafObject, isNotNull);
    ensure(leafObject, isNotNull);
    ensure(propertyPath, isNotNull); // TODO: is empty valid
    ensure(rootObject, isNotNull);
  }

  String toString() => message;

  /*
   Java's properties (some of them)
ConstraintDescriptor<?>  getConstraintDescriptor()
Object  getInvalidValue()
Returns the value failing to pass the constraint.
Object  getLeafBean()
Returns: the bean instance the constraint is applied on if it is a bean constraint the bean instance hosting the property the constraint is applied on if it is a property constraint null when the ConstraintViolation is returned after calling Validator.validateValue(Class, String, Object, Class[]) the object the method is executed on if it is a method parameter, cross-parameter or return value constraint null if it is a constructor parameter or cross-parameter constraint the object the constructor has created if it is a constructor return value constraint
String  getMessage()
String  getMessageTemplate()
Path  getPropertyPath()
T getRootBean()
Returns the root bean being validated.

   */
}

typedef Matcher MatcherCreator();





class _ConstraintFailureHandler implements FailureHandler {
  final ConstraintValidationContext context;
  _ConstraintFailureHandler(this.context);

  void fail(String reason) {
    context.addViolation(reason: reason);
//    violations.add(new ConstraintViolation(reason));
  }

  void failMatch(actual, Matcher matcher, String reason,
      Map matchState, bool verbose) {
    fail(_defaultErrorFormatter(actual, matcher, reason, matchState, verbose));
  }
}

// copied from matcher/expect.dart
String _defaultErrorFormatter(actual, Matcher matcher, String reason,
    Map matchState, bool verbose) {
  var description = new StringDescription();
  description.add('Expected: ').addDescriptionOf(matcher).add('\n');
  description.add('  Actual: ').addDescriptionOf(actual).add('\n');

  var mismatchDescription = new StringDescription();
  matcher.describeMismatch(actual, mismatchDescription, matchState, verbose);

  if (mismatchDescription.length > 0) {
    description.add('   Which: ${mismatchDescription}\n');
  }
  if (reason != null) {
    description.add(reason).add('\n');
  }
  return description.toString();
}

class _MatcherValidator implements Function {
  void call(dynamic fieldValue, Ensure constraint,
      ConstraintValidationContext context) {
    final matcher = constraint.matcher is MatcherCreator ? constraint.matcher()
        :constraint.matcher;

    if (!(matcher is Matcher)) {
      throw new ArgumentError('matcher must be either a Matcher or a function'
          ' that returns a Matcher');
    }
    final handler = new _ConstraintFailureHandler(context);
    expect(fieldValue, matcher, failureHandler: handler);
  }


  const _MatcherValidator();
}

