// Copyright (c) 2014, The Constrain project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

library constraint.validate.context;

import 'package:constrain/metadata.dart';
import 'package:constrain/constraint.dart';
import 'preconditions.dart';
import 'dart:mirrors';

class _ConstraintViolationContainer {
  final Set violations = new Set<ConstraintViolation>();

  void addViolation(ConstraintViolation violation) {
    violations.add(violation);
  }
}

abstract class ValidationContext  {
  final InstanceMirror instanceMirror;
  final ValidationContext parent;
  final TypeDescriptorResolver resolver;
  final _ConstraintViolationContainer violations;
  dynamic get object => instanceMirror.reflectee;

  ValidationContext(this.instanceMirror, this.resolver, this.violations,
      [this.parent]) {
    ensure(instanceMirror, isNotNull);
    ensure(resolver, isNotNull);
    ensure(violations, isNotNull);
  }

  void validate();

  Iterable<Symbol> get path;

  TypeValidationContext get root => parent == null ? this : parent.root;
}

class TypeValidationContext extends ValidationContext {

  TypeValidationContext(InstanceMirror instanceMirror,
                         TypeDescriptorResolver resolver,
                         _ConstraintViolationContainer violations,
                         [ValidationContext parent])
      : super(instanceMirror, resolver, violations, parent);

  TypeValidationContext.root(Object object, TypeDescriptorResolver resolver)
      : this(reflect(object), resolver, new _ConstraintViolationContainer());

  void validate() {
    if (object != null) {
      final typeDescriptor = resolver.resolveFor(object.runtimeType);
      final members = typeDescriptor.memberDescriptors;
      members.forEach(_validateMember);
    }
  }

  void _validateMember(MemberDescriptor member) {
    final memberContext = pushMember(member);
    memberContext.validate();
  }

  MemberValidationContext pushMember(MemberDescriptor member) {
    return new MemberValidationContext(instanceMirror.getField(member.name),
        member, resolver, violations, this);
  }

  Iterable<Symbol> get path =>
      parent == null ? [] : parent.path;

  Set<ConstraintViolation> build() =>
      violations.violations;

}


class MemberValidationContext extends ValidationContext {
  final MemberDescriptor member;

  MemberValidationContext(InstanceMirror instanceMirror,
                          this.member,
                          TypeDescriptorResolver resolver,
                          _ConstraintViolationContainer violations,
                          TypeValidationContext parent)
      : super(instanceMirror, resolver, violations, parent) {
    ensure(parent, isNotNull);
    ensure(member, isNotNull);
  }

  @override
  void validate() {
    member.constraintDescriptors.forEach(_validateConstraint);
    // TODO: now validate the fields object (i.e. a type context again)
  }

  void _validateConstraint(ConstraintDescriptor cd) {
    final constraintContext =
        new MemberConstraintValidationContext(cd, instanceMirror, member,
        resolver, violations, parent); // TODO: this or parent?

    constraintContext.validate();
  }

  Iterable<Symbol> get path =>
      parent == null ? [] : new List.from(parent.path)..add(member.name);

}

class MemberConstraintValidationContext extends MemberValidationContext
    implements ConstraintValidationContext {
  final ConstraintDescriptor constraintDescriptor;

  MemberConstraintValidationContext(this.constraintDescriptor,
        InstanceMirror instanceMirror, MemberDescriptor member,
        TypeDescriptorResolver resolver,
        _ConstraintViolationContainer violations,
        TypeValidationContext parent)
      : super(instanceMirror, member, resolver, violations, parent) {
    ensure(constraintDescriptor, isNotNull);
  }

  @override
  void validate() {
    constraintDescriptor.constraint.validate(instanceMirror.reflectee, this);
  }

  void addViolation({String reason}) {
    violations.addViolation(new ConstraintViolation(constraintDescriptor,
        object, parent.object, path, root.object, reason: reason));
  }
}
